/*-------------------------------------------------------------------*/
/* CBSHOW.C       Juha Kivijrvi                                     */
/*                Timo Kaukoranta                                    */
/*                Pasi Franti                                        */
/*                                                                   */
/* - Show content of the training set file and the codebook file.    */
/* - Added  #include "interfc.h", to make it compile. -VH            */
/*-------------------------------------------------------------------*/

#define ProgName       "CBSHOW"
#define VersionNumber  "V.0.17"
#define LastUpdated    "9.11.00"

/* ----------------------------------------------------------------- */


#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <math.h>

#define  FACTFILE  "cbshow.fac"

#include "parametr.c"
#include "cb.h"
#include "file.h"
#include "memctrl.h"
#include "owntypes.h"
#include "interfc.h"

/*-----------------------------  M i s c  ----------------------------*/

#define max(a,b)     ( (a) > (b) ? (a) : (b) )
#define max3(a,b,c)  ( max( max(a,b), c ) )

#define strlength(a) ( (a) == NULL ? 0 : strlen(a) )

#define width(a)   ( (a) > 0  ? (int) log10(a) + 1 : \
                     (a) == 0 ? 1 : (int) log10(-a) +2 )

/*============================  P R I N T  ===============================*/


void PrintInfo(void)
{
  printf("\n%s %s %s\n", ProgName, VersionNumber, LastUpdated);
  printf("Prints a content of training set file and codebook file.\n");
  printf("Usage: %s [%coption] <set-file>\n", ProgName, OPTION_SYMBOL);
  printf("Input file must be TR-file or CB-file.\n");
  printf("For example: %s set_of_4\n", ProgName);
  printf("\n  Options:\n");
  PrintOptions();
}


/*-------------------------------------------------------------------*/


static void PrintOperatingInfo(char* InFileName, TRAININGSET* TS)
{
  if( Value(ShowProgress) >= 1 )
    {
    printf("\n%s %s %s\n", ProgName, VersionNumber, LastUpdated);
    printf("Input file:        %s\n", InFileName);
    printf("Version:           %s\n", TS->Versionstr);
    printf("BlockSizeX:        %i\n", TS->BlockSizeX);
    printf("BlockSizeY:        %i\n", TS->BlockSizeY);
    printf("Booksize:          %i\n", BookSize(TS));
    printf("Total frequency:   %i\n", TotalFreq(TS));
    printf("Bytes per element: %i\n", TS->BytesPerElement);
    printf("MinValue:          %i\n", TS->MinValue);
    printf("MaxValue:          %i\n", TS->MaxValue);
    printf("Preprocessing:     %i\n", TS->Preprocessing);
    if (strlen(TS->GenerationMethod) > 0)
      printf("Generation method: %s\n", TS->GenerationMethod);
    printf("\n");
/*
    PrintSelectedOptions();
    printf("\n");
*/
    }
}


/*=======================  E N T R O P Y  ============================*/


static double Entropy(double p)
{
  return( p > 0.0 ? -p * log(p) / log(2) : 0.0 );
}


/*-------------------------------------------------------------------*/


static double CalculateEntropy(TRAININGSET *TS)
{
  double  entropy = 0, prop;
  int     i;

  for( i = 0; i < BookSize(TS); i++ )
    {
    prop = (double) VectorFreq(TS,i) / (double) TotalFreq(TS);
    entropy += Entropy(prop);
    }

  return( entropy );
}


/*=======================  P R O C E S S  ==========================*/


static void CalculateMaxValues(TRAININGSET* TS,
                               int* maxfreq, 
                               int* maxelement,
                               int* maxnamewidth)
{
  int i, j;

  *maxfreq = *maxelement = *maxnamewidth = 0;

  for ( i = 0; i < BookSize(TS); i++ )
    {
    if ( VectorFreq(TS, i) > *maxfreq )  
      *maxfreq = VectorFreq(TS, i);
    if ( strlength(VectorName(TS, i)) > *maxnamewidth )
      *maxnamewidth = strlength(VectorName(TS, i));

    for ( j = 0; j < VectorSize(TS); j++ )
      {
      if ( VectorScalar(TS, i, j) > *maxelement )
        *maxelement = VectorScalar(TS, i, j);
      }
    }  
}



/*-------------------------------------------------------------------*/


static void LinePrintNode(BOOKNODE* node, 
                          int       vectorsize, 
                          int       namewidth,
                          int       elementwidth, 
                          int       freqwidth)
{
  int i;

  if (namewidth > 0)
    {
    if (strlength(node->name) > 0)
      printf("%-*s  ", namewidth, node->name);
    else
      for ( i = 0; i < namewidth + 2; i++ ) printf(" ");
    }

  for( i = 0; i < vectorsize; i++ )
    {
    printf("%*d", elementwidth, node->vector[i]);
    if ( ! (Value(NoBlanks)) )
      printf(" ");
    }
  printf(" (%*d)\n", freqwidth, node->freq);
}  


/*-------------------------------------------------------------------*/


static void PrintTitle(int i, 
                       int VectorsPerRow, 
                       int vectorwidth, 
                       int indexwidth, 
                       int freqwidth, 
                       TRAININGSET *TS)
{
  int k;
  int spaces = vectorwidth - indexwidth - freqwidth - 1;
  int sp;

  for(k = 0; k < VectorsPerRow; k++)
    {
    if( i + k < BookSize(TS))
      {
      printf("%*u/%-*u", indexwidth, i + k, freqwidth, VectorFreq(TS,i + k));
      if( k < VectorsPerRow - 1 )
        {
        for( sp = 0; sp < spaces; sp++)
          {
          putchar(' ');
          }
        }
      }
    }
  putchar('\n');
}


/*-------------------------------------------------------------------*/


static void PrintNames(int i, 
                       int VectorsPerRow, 
                       int vectorwidth,
                       TRAININGSET *TS)
{
  int k;
  int sp;

  for(k = 0; k < VectorsPerRow; k++)
    {
    if( i + k < BookSize(TS))
      {
      if ( strlength(VectorName(TS, i+k)) > 0 )
        {
        printf("%s", VectorName(TS, i+k));
        if ( k < VectorsPerRow - 1)
          {
          for ( sp = 0; sp < vectorwidth - strlength(VectorName(TS, i+k)); sp++ )
            {
            putchar(' ');
            }
          }
        } 
      else
        {
        if ( k < VectorsPerRow - 1)
          {
          for ( sp = 0; sp < vectorwidth; sp++ )
            {
            putchar(' ');
            }
          }
        }
      }
    }
  putchar('\n');
}


/*-------------------------------------------------------------------*/


static void PrintVectors(int i, 
                         int VectorsPerRow, 
                         int vectorwidth,
                         int elementwidth,
                         TRAININGSET *TS)
{
  int k, row, col;
  int spaces = vectorwidth - TS->BlockSizeX * (elementwidth + 1) - 2;
  int sp;

  for( row = 0; row < TS->BlockSizeY; row++ )
    {
    for( k = 0; k < VectorsPerRow; k++ )
      {
      if( i + k < BookSize(TS) )
        {
        for(col = 0; col < TS->BlockSizeX; col++)
          {
          printf("%*i", elementwidth, 
                        VectorScalar(TS,i + k,row * TS->BlockSizeX + col));
          if( k < VectorsPerRow - 1 ||
              col < TS->BlockSizeX - 1 )
            {
            putchar(' ');
            }
          }
        for ( sp = 0; sp < spaces; sp++ )
          {
          putchar(' ');
          }

        if( k < VectorsPerRow - 1 && i + k < BookSize(TS) - 1)
          {
          printf("; ");
          }
        }
      }
    putchar('\n');
    }
}


/*-------------------------------------------------------------------*/


static void PrintTrainingSet(TRAININGSET* TS, 
                             CODEBOOK*    CB,
                             PARTITIONING* P,
                             CBFILETYPE mode)
{
  int         VectorWidth;
  int         VectorsPerRow;
  int         i,j,k;
  double      entropy;
  int         maxfreq = 0;
  int         maxelement = 0;
  int         maxnamewidth = 0;
  
  CalculateMaxValues(TS, &maxfreq, &maxelement, &maxnamewidth);

  VectorWidth = 1 + max3( (1 + width(maxelement)) * TS->BlockSizeX + 1,
                           width(BookSize(TS)-1) + width(maxfreq) + 2, 
                           maxnamewidth + 1);

  VectorsPerRow = max(1, 80 / VectorWidth);

  if( Value(CalculateEntropy) == YES )
    {
    entropy = CalculateEntropy(TS);
    printf("Entropy of the set = %6.3f (bits/vector) \n", entropy);
    }

  if( Value(ShowProgress) >= 2 )
    {
    printf("\n");

    if ( mode == TSFILE )
      {
      if ( Value(VectorFormat) == MatrixFormat || 
          (Value(VectorFormat) == 0 && TS->BlockSizeY > 1 ))
        {
        for( i = 0; i < BookSize(TS); i += VectorsPerRow )
          {
          PrintTitle(i, 
                     VectorsPerRow, 
                     VectorWidth,
                     width(BookSize(TS)-1),
                     width(maxfreq), 
                     TS);
          if (maxnamewidth > 0)
            PrintNames(i,
                       VectorsPerRow,
                       VectorWidth,
                       TS);
          PrintVectors(i, 
                       VectorsPerRow, 
                       VectorWidth,
                       width(maxelement),
                       TS);
          putchar('\n');
          }
        }
      else         /* line format */
        {
        for( i = 0; i < BookSize(TS); i++ )
          {
          LinePrintNode(&Node(TS,i), 
                        VectorSize(TS), 
                        maxnamewidth,
                        width(maxelement), 
                        width(maxfreq));
          }
        }
      }
    else           /* partitioning print */
      {
      for ( i = 0; i < BookSize(CB); i++)
        {
        printf("Partition %d:\n", i+1);
        printf("   centroid: ");
        for ( j = 0; j < maxnamewidth + width(BookSize(CB)) - 6; j++ )
          printf(" "); 
        LinePrintNode(&Node(CB,i), 
                      VectorSize(TS), 
                      0,
                      width(maxelement), 
                      width(maxfreq));
        for ( j = FirstVector(P,i); !(EndOfPartition(j)); j = NextVector(P,j) )
          {
          printf("   %*d  ", width(BookSize(CB)), i+1);
          for ( k = 0; k < 6 - maxnamewidth - width(BookSize(CB)) + 
                          ((maxnamewidth == 0) ? 2 : 0); k++ )
            printf(" "); 
          LinePrintNode(&Node(TS, j), 
                        VectorSize(TS), 
                        maxnamewidth,
                        width(maxelement), 
                        width(maxfreq));
          }
        }                  
      }
    }
}


/*=========================  T Y P E S  ===============================*/


static CBFILETYPE CheckTypes(CBFILETYPE ft1, CBFILETYPE ft2)
{
  if ( ft2 == NOTFOUND )
    {
    if ( ft1 == TSFILE || ft1 == CBFILE) return TSFILE;
    else 
      {
      printf("ERROR: first file not found or a partitioning file\n");
      exit(-1);
      }
    }
  else
    {
    if ( ft1 == TSFILE )
      {
      if ( ft2 == CBFILE ) return CBFILE;
      if ( ft2 == PAFILE ) return PAFILE;
      printf("ERROR: second file not found or a training set file\n");
      exit(-1);
      }
    else
      {
      printf("ERROR: first file not a training set file\n");
      exit(-1);
      }
    }  
}



/*=====================  I N I T I A L I Z E  ==========================*/


static void Initialize(void)
{
}


/*=====================  S H U T D O W N  ============================*/


static void ShutDown(TRAININGSET* TS,
                     CODEBOOK*    CB,
                     PARTITIONING* P,
                     CBFILETYPE mode)
{
  FreeCodebook(TS);

  if (mode != TSFILE)
    {
    FreeCodebook(CB);
    FreePartitioning(P);
    }

}


/*===========================  M A I N  ==============================*/


int main(int argc, char* argv[])
{
  char        FileName1[80];
  char        FileName2[80];
  CBFILETYPE  ft1, ft2, mode;

  TRAININGSET TS;
  CODEBOOK    CB;
  PARTITIONING P;

  ParameterInfo paraminfo[2] =  { { FileName1, "", 0, INFILE },
                                  { FileName2, "", 1, INFILE } };
                                
  ParseParameters( argc, argv, 2, paraminfo);

  ft1 = DetermineCBFileType(FileName1);
  ft2 = DetermineCBFileType(FileName2);

  if ( ft2 == NOTFOUND && strlength(FileName2) > 0 )
    {
    printf("ERROR: file not found.\n");
    exit(-1);
    }

  mode = CheckTypes(ft1, ft2);

  ReadTrainingSet(FileName1, &TS);
  if ( mode == CBFILE ) 
    {
    ReadCodebook(FileName2, &CB);
    CreateNewPartitioning(&P, &TS, BookSize(&CB));
    GenerateOptimalPartitioning(&TS, &CB, &P);
    }
  if ( mode == PAFILE ) 
    {
    ReadPartitioning(FileName2, &P, &TS);
    CreateNewCodebook(&CB, PartitionCount(&P), &TS);
    GenerateOptimalCodebook(&TS, &CB, &P);
    }

  Initialize();

  PrintOperatingInfo(FileName1,&TS);
  PrintTrainingSet(&TS, &CB, &P, mode);

  ShutDown(&TS, &CB, &P, mode);
  checkmemory();
  
  return( 0 );
}

